import os

from .file_loaders import load_yaml, load_str, load_lst, load_hcl, load_toml
from .special_loaders import load_x

EXT_TO_LOADER = {
    'yml': load_yaml,
    'yaml': load_yaml,
    'json': load_yaml,
    'text': load_str,
    'txt': load_str,
    '': load_str,
    'lst': load_lst,
    'hcl': load_hcl,
    'toml': load_toml,
    'tml': load_toml,
    'x': load_x,
}


def load_file(path):
    if not os.path.exists(path):
        print(f'[ERROR] {path} not found, ignored...')
        return {}

    ext = path.split('.')[-1] if '.' in path else ''

    return EXT_TO_LOADER.get(ext, load_str)(path)



import traceback


def safe(error_msg='Could not load', default_value=None):
    def wrapper(function):
        def _(path):
            try:
                return function(path)
            except: # pylint: disable=bare-except
                print(f'[ERROR] {error_msg}: {path}')
                traceback.print_exc()
                return default_value
        return _
    return wrapper


def with_open(function):
    def wrapped(path):
        with open(path, 'r', encoding='utf-8') as file_:
            return function(file_)
    return wrapped


def path_to_name(path):
    return path.split('/')[-1].split('.')[0]



import yaml

import toml
import hcl2 as hcl

from .utils import with_open, safe


@with_open
@safe(error_msg='Could not parse yaml')
def load_yaml(file_):
    docs = list(yaml.unsafe_load_all(file_))
    return docs[0] if len(docs) == 1 else docs


@with_open
@safe(error_msg='Could not parse HCL (v2)')
def load_hcl(file_):
    return hcl.load(file_)


@with_open
@safe(error_msg='Could not parse TOML')
def load_toml(file_):
    return toml.loads(file_.read())


@with_open
@safe(error_msg='Could not read', default_value='')
def load_str(file_):
    return file_.read()


@safe(error_msg='Could not parse list', default_value=[])
def load_lst(path):
    return load_str(path).strip().replace('\r', '').split('\n')



import os
import copy

from ..exceptions import GaelinaUnmergeableException
from .file_loader import load_file
from .utils import path_to_name

ELEMENTS_TO_IGNORE = '.git', 'filters.py', '__pycache__'


def load_any(path):
    return load_dir(path) if os.path.isdir(path) else load_file(path)


def load_dir(dir_path):
    data = None
    for dir_element_path in os.listdir(dir_path):
        if dir_element_path in ELEMENTS_TO_IGNORE:
            continue
        sub_data = load_any(dir_path + '/' + dir_element_path)
        if dir_element_path.startswith('_'):
            data = merged_data(data, sub_data)
        else:
            data = merged_data(data, {path_to_name(dir_element_path): sub_data})
    return data


def merged_data(d1, d2):
    if d1 is None and d2 is not None:
        return d2
    type_match = (type(d1), type(d2))
    if type_match == (list, list):
        return d1 + d2
    if type_match == (dict, dict):
        return merged_dicts(d1, d2)
    raise GaelinaUnmergeableException(f'Unmergeable data {d1} and {d2}.')


def merged_dicts(d1, *dicts):
    res = copy.deepcopy(d1 or {})
    for d2 in dicts:
        rec_add_dict(res, d2 or {}, [])
    return res


def rec_add_dict(dest, src, path):
    for k, v in src.items():
        if k not in dest:
            dest[k] = copy.deepcopy(v)
        else:
            type_match = (type(dest[k]), type(v))
            if type_match == (list, list):
                dest[k] += v
            elif type_match == (dict, dict):
                rec_add_dict(dest[k], v, path + [k])
            else:
                raise GaelinaUnmergeableException(f'Trying to replace existing value for key {k} at {" -> ".join(path)}.')



import jinja2

from .loaders import load_any, load_filters
from .exceptions import GaelinaRenderingException

MAX_ITERATIONS = 200


def gaelina(path):
    source = Source(path)
    if isinstance(source.data, dict):
        return GaelinaDict(source.data, source)
    return source.data


def gaelina_value(value, source):
    if isinstance(value, dict):
        return GaelinaDict(value, source)
    if isinstance(value, list):
        return GaelinaList(value, source)
    if isinstance(value, str) and '{{' in value:
        return g_render(value, source)
    return value


def g_render(str_, source):
    render = str_
    for _ in range(MAX_ITERATIONS):
        new_render = source.render(render)
        if new_render == render:
            return render
        if not isinstance(new_render, str):
            return gaelina_value(new_render, source)
        render = new_render
    raise GaelinaRenderingException(f'Failed to render template: {str_}')


class Source:
    def __init__(self, path):
        self.data = load_any(path)
        self.filters = load_filters(path, self.data)
        self.env = jinja2.Environment() # nosec
        for filter_name, filter_ in self.filters.items():
            self.env.filters[filter_name] = filter_

    def consider_a_value(self, str_):
        return str_.startswith('{{') and str_.endswith('}}') and '{{' not in str_[2:]

    def render(self, str_):
        if self.consider_a_value(str_):
            expression = str_[2:-2].strip()
            return self.env.compile_expression(expression)(self.data)

        # Somehow trailing line feed are removed by jinja2 readding them
        trail = '\n' if str_.endswith('\n') else ''

        return self.env.from_string(str_).render(**self.data) + trail


class GaelinaDict(dict):
    def __init__(self, value, source):
        super().__init__(value)
        self.source = source

    def __getitem__(self, key):
        value = super().__getitem__(key)
        return gaelina_value(value, self.source)

    def get(self, key, default_value):
        value = super().get(key, default_value)
        return gaelina_value(value, self.source)


class GaelinaList(list):
    def __init__(self, value, source):
        super().__init__(value)
        self.source = source

    def __getitem__(self, index):
        value = super().__getitem__(index)
        if isinstance(index, slice):
            return GaelinaList(value, self.source)
        return gaelina_value(value, self.source)



import os

from importlib.machinery import SourceFileLoader
from importlib.util import spec_from_loader, module_from_spec


def load_filters(path, context):
    module = load_filter_module(f'{path}/filters.py')
    filters = getattr(module, 'get_filters', lambda _: {})(context)
    if len(filters) == 0:
        print('No filters loaded.')
    return filters


def load_filter_module(path):
    if not os.path.exists(path):
        print('No filters file found.')
        return None

    module_name = 'custom_filters'
    loader = SourceFileLoader(module_name, path)
    spec = spec_from_loader(module_name, loader)
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module



import requests

from ..utils import safe


API_URL = 'https://publish.twitter.com/oembed'


class Xception(Exception):
    pass


@safe(error_msg='Could not load X file')
def load_x(path):
    with open(path, 'r', encoding='utf-8') as file_:
        lines = [line for line in file_.read().split('\n') if line]
        if len(lines) > 3:
            raise Xception('X file has too many lines (>2)')
        if len(lines) == 0:
            raise Xception('Empty X file')
        if len(lines) == 2:
            return {'link': lines[0], 'embedded': lines[1]}

    embedded = get_embedded_tweet(lines[0])
    update_x_file(path, lines[0], embedded)

    return {'link': lines[0], 'embedded': embedded}


def get_embedded_tweet(tweet_link):
    response = requests.get(API_URL, params={'url': tweet_link, 'omit_script': 'false'}, timeout=30)
    return response.json()['html']


def update_x_file(path, link, embedded):
    with open(path, 'w', encoding='utf-8') as file_:
        file_.write(f'{link}\n{embedded}')



import os
import re
from setuptools import setup, find_packages

tag = os.environ.get('CI_COMMIT_TAG')
branch = os.environ.get('CI_COMMIT_BRANCH')
version = tag or (branch if branch and re.match(r'\d+\.\d+', branch) else '0.0')

if __name__ == '__main__':
    setup(
        version=version,
        long_description=open('README.md').read() + '\n\n\n' + open('CHANGELOG.md').read(),
        long_description_content_type='text/markdown',
        packages=find_packages(where='src', exclude=['tests']),
    )
